
/*
    VPPP virtual PPP connection over TCP.

    Copyright (C) 1998,1999  Maxim Krasnyansky <max@uznet.net>

    Idea is based on Sam Lantinga's PPPTCP package, but code was 
    completely written from the scratch.

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
*/

/*
 * Version: 2.1 03/06/1999 Maxim Krasnyansky <max_mk@yahoo.com>
 */ 

#include <fcntl.h>
#include <stdlib.h>
#include <signal.h>
#include <stdlib.h>
#include <stdio.h>
#include <syslog.h>
#include <strings.h>
#include <stdarg.h>
#include <unistd.h>

#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/types.h>

#include <sys/socket.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>

#include <linux/if_slip.h>

#include <errno.h>

#include "vppp.h"

extern int errno;

extern char *proc_name;

/* Allocate psevdo tty 
 * returns master side fd and copies slave name to first arg
 */  
int getpty(char *sl_name){
	char ptyname[] = "/dev/ptyXY";
	char ch[] = "pqrstuvwxyzabcde";
	char digit[] = "0123456789abcdef";
	int l, m;
	int mr_fd;
	
	for(l=0; ch[l]; l++ ){
		for(m=0; digit[m]; m++ ){
			ptyname[8] = ch[l];
			ptyname[9] = digit[m];
			/* Open the master */
			if((mr_fd=open(ptyname, O_RDWR)) < 0)
				continue;
			/* Check the slave */
			ptyname[5] = 't';
			if( (access(ptyname, R_OK | W_OK)) < 0 ){
				close(mr_fd);
			 	ptyname[5] = 'p';
				continue;
			}
			strcpy(sl_name,ptyname);
			return mr_fd;
		}
	}
	return -1;
}



int start_ppp(int rt_fd, char *host, int speed, char *ppp_arg){
     int mr_fd,sl_fd;
     int pid,opt;
     char pty[10];

     if( (mr_fd=getpty(pty)) < 0){
	syslog(LOG_ERR,"Can not allocate psevdo tty");
	return -1;
     }
	
     /* Fork and run the program on the pseudo-tty */
     switch( (pid=fork()) ){
	case -1:
		syslog(LOG_ERR,"Couldn't fork()");
		return -1;
	case 0:
		setsid();
		/* Open pty slave -- it becomes controlling terminal */
		if( (sl_fd = open(pty, O_RDWR, 0)) < 0){
		   syslog(LOG_ERR,"Couldn't open slave pty.");
		   return -1;
		}
		/* Close master side of pty and route fds to slave */
		close(mr_fd);
		close(0); dup(sl_fd);
		close(1); dup(sl_fd);
		close(2);
		close(sl_fd);
		exec_pppd(ppp_arg);

	default:
		set_title("%s: host %s pppd pid <%d>", proc_name, host, pid);

	        opt=1;
	        setsockopt(rt_fd,SOL_SOCKET,SO_KEEPALIVE,&opt,sizeof(opt));

	        opt=1;
	        setsockopt(rt_fd,IPPROTO_TCP,TCP_NODELAY,&opt,sizeof(opt));

		switch(speed){
		   case 0:
		   	link_fd_std(rt_fd,mr_fd);
			break;
		   case -1:
		   	link_fd_cmp(rt_fd,mr_fd);
			break;
  		   default: 
		   	link_fd_shp(rt_fd,mr_fd,speed);
			break;
		}

		close(mr_fd);
	        return 0;
	}
}

void exec_pppd(char *arg){
     char *argv[50],*str;
     int i=2;	

     argv[0]="pppd";	
     if(arg){
       argv[1]=strtok(arg," ");     
       while( (str=strtok(NULL," ")) && i < 50)
	  argv[i++]=str;
       argv[i]=NULL;
     } else
       argv[1]=NULL;

     execv(PPPD,argv);
}
  	
extern char **environ;
char	*title_start;	/* start of the proc title space */
char	*title_end;     /* end of the proc title space */
int	title_size;


void init_title(int argc,char *argv[], char *envp[])
{
	int i;

	/*
	**  Move the environment so settitle can use the space at
	**  the top of memory.
	*/

	for (i = 0; envp[i]; i++);

	environ = (char **) malloc(sizeof (char *) * (i + 1));

	for(i = 0; envp[i]; i++)
	   environ[i] = strdup(envp[i]);
	environ[i] = NULL;

	/*
	**  Save start and extent of argv for set_title.
	*/

	title_start = argv[0];

	/*
	**  Determine how much space we can use for set_title.  
	**  Use all contiguous argv and envp pointers starting at argv[0]
 	*/
	for(i=0; i<argc; i++)
	   if( !i || title_end == argv[i])
	     title_end = argv[i] + strlen(argv[i]) + 1;

	for(i=0; envp[i]; i++)
  	   if( title_end == envp[i] )
	     title_end = envp[i] + strlen(envp[i]) + 1;

	title_size=title_end-title_start;
}

void set_title(const char *fmt, ...)
{
	char buf[255];
	va_list ap;

	memset(title_start,0,title_size);

	/* print the argument string */
	va_start(ap, fmt);
	vsprintf(buf, fmt, ap);
	va_end(ap);

	if( strlen(buf) > title_size - 1)
	  buf[title_size - 1] = '\0';

	strcpy(title_start, buf);
}
    
int write_n(int fd, char *buf, int len)
{
	int t=0, w;

	do {
 	  if( (w = write(fd, buf, len)) < 0 )
	    return -1;
	  len -= w;
	  buf += w;
	  t += w;
	} while(len > 0);

	return(t);
}

/* Connect with timeout */

int connect_t(int s, struct sockaddr *svr, time_t timeout) 
{
	int sock_flags;
	fd_set fdset;
	struct timeval tv;

	tv.tv_usec=0; tv.tv_sec=timeout;

	sock_flags=fcntl(s,F_GETFL);
	if( fcntl(s,F_SETFL,O_NONBLOCK) < 0 )
	  return -1;

	if( connect(s,svr,sizeof(struct sockaddr)) < 0 && errno != EINPROGRESS)
	  return -1;

	FD_ZERO(&fdset);
	FD_SET(s,&fdset);
	if( select(s+1,NULL,&fdset,NULL,&tv) > 0 ){
	  int l=sizeof(errno);	 
 	  errno=0;
 	  getsockopt(s,SOL_SOCKET,SO_ERROR,&errno,&l);
	} else
	  errno=ETIMEDOUT;  	

	fcntl(s,F_SETFL,sock_flags); 

 	if(errno)
	  return -1;

	return 0;
}	

/* Read with timeout */

int read_t(int fd, void *buf, size_t count, time_t timeout) 
{
	fd_set fdset;
	struct timeval tv;

	tv.tv_usec=0; tv.tv_sec=timeout;

	FD_ZERO(&fdset);
	FD_SET(fd,&fdset);
	if( select(fd+1,&fdset,NULL,NULL,&tv) <= 0)
	   return -1;

	return read(fd,buf,count);
}
